# validated: 2018-01-04 DV 1f4822f33278 edu/wpi/first/wpilibj/drive/DifferentialDrive.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2018. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import math

import hal
from .robotdrivebase import RobotDriveBase

__all__ = ["DifferentialDrive"]


class DifferentialDrive(RobotDriveBase):
    """
    A class for driving differential drive/skid-steer drive platforms such as the Kit of Parts drive
    base, "tank drive", or West Coast Drive.

    These drive bases typically have drop-center / skid-steer with two or more wheels per side
    (e.g., 6WD or 8WD).

    This class takes a :class:`.SpeedController` per side. For four and six motor
    drivetrains, construct and pass in :class:`.SpeedControllerGroup` instances as follows.

    Four motor drivetrain::

        def robotInit(self):
            self.frontLeft = wpilib.Spark(1)
            self.rearLeft = wpilib.Spark(2)
            self.left = wpilib.SpeedControllerGroup(self.frontLeft, self.rearLeft)

            self.frontRight = wpilib.Spark(3)
            self.rearRight = wpilib.Spark(4)
            self.right = wpilib.SpeedControllerGroup(self.frontRight, self.rearRight)

            self.drive = DifferentialDrive(self.left, self.right)

    Six motor drivetrain::

        def robotInit(self):
            self.frontLeft = wpilib.Spark(1)
            self.midLeft = wpilib.Spark(2)
            self.rearLeft = wpilib.Spark(3)
            self.left = wpilib.SpeedControllerGroup(self.frontLeft, self.midLeft, self.rearLeft)

            self.frontRight = wpilib.Spark(4)
            self.midRight = wpilib.Spark(5)
            self.rearRight = wpilib.Spark(6)
            self.right = wpilib.SpeedControllerGroup(self.frontRight, self.midRight, self.rearRight)

            self.drive = DifferentialDrive(self.left, self.right)

    A differential drive robot has left and right wheels separated by an arbitrary width.

    Drive base diagram::

        |_______|
        | |   | |
          |   |
        |_|___|_|
        |       |


    Each ``drive()`` function provides different inverse kinematic relations for a differential drive
    robot. Motor outputs for the right side are negated, so motor direction inversion by the user is
    usually unnecessary.

    This library uses the NED axes convention (North-East-Down as external reference in the world
    frame): http://www.nuclearprojects.com/ins/images/axis_big.png.

    The positive X axis points ahead, the positive Y axis points right, and the positive Z axis
    points down. Rotations follow the right-hand rule, so clockwise rotation around the Z axis is
    positive.

    Inputs smaller than :data:`.RobotDriveBase.kDefaultDeadband` will be set to
    0, and larger values will be scaled so that the full range is still used.
    This deadband value can be changed with :meth:`~.RobotDriveBase.setDeadband`.

    .. note:: RobotDrive porting guide:

        :meth:`.tankDrive` is equivalent to
        :meth:`.RobotDrive.tankDrive` if a deadband of 0 is used.

        :meth:`.arcadeDrive` is equivalent to
        :meth:`.RobotDrive.arcadeDrive` if a deadband of 0 is used
        and the rotation input is inverted (i.e ``arcadeDrive(y, -rotation)``)

        :meth:`.curvatureDrive` is similar in concept to
        :meth:`.RobotDrive.drive` with the addition of a quick turn
        mode. However, it is not designed to give exactly the same response.
    """

    kDefaultQuickStopThreshold = .2
    kDefaultQuickStopAlpha = .1

    instances = 0

    def __init__(self, leftMotor, rightMotor):
        """Constructor for DifferentialDrive.

        .. note:: To pass multiple motors per side, use a :class:`.SpeedControllerGroup`.
                  If a motor needs to be inverted, do so before passing it in.

        :param leftMotor: Left motor(s)
        :param rightMotor: Right motor(s)
        :type leftMotor: :class:`.SpeedController`
        :type rightMotor: :class:`.SpeedController`
        """
        super().__init__()

        self.leftMotor = leftMotor
        self.rightMotor = rightMotor

        self.quickStopThreshold = self.kDefaultQuickStopThreshold
        self.quickStopAlpha = self.kDefaultQuickStopAlpha
        self.quickStopAccumulator = 0.0
        self.reported = False

        self.addChild(self.leftMotor)
        self.addChild(self.rightMotor)
        DifferentialDrive.instances += 1
        self.setName("DifferentialDrive", self.instances)

    def arcadeDrive(self, xSpeed, zRotation, squaredInputs=True):
        """Arcade drive method for differential drive platform.

        :param xSpeed: The robot's speed along the X axis `[-1.0..1.0]`. Forward is positive
        :param zRotation: The robot's zRotation rate around the Z axis `[-1.0..1.0]`. Clockwise is positive
        :param squaredInputs: If set, decreases the sensitivity at low speeds.
        """

        if not self.reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       2,
                       hal.UsageReporting.kRobotDrive_ArcadeStandard)
            self.reported = True

        xSpeed = RobotDriveBase.limit(xSpeed)
        xSpeed = RobotDriveBase.applyDeadband(xSpeed, self.deadband)

        zRotation = RobotDriveBase.limit(zRotation)
        zRotation = RobotDriveBase.applyDeadband(zRotation, self.deadband)

        if squaredInputs:
            # Square the inputs (while preserving the sign) to increase fine
            # control while permitting full power.
            xSpeed = math.copysign(xSpeed * xSpeed, xSpeed)
            zRotation = math.copysign(zRotation * zRotation, zRotation)

        maxInput = math.copysign(max(abs(xSpeed), abs(zRotation)), xSpeed)

        if xSpeed >= 0.0:
            if zRotation >= 0.0:
                leftMotorSpeed = maxInput
                rightMotorSpeed = xSpeed - zRotation
            else:
                leftMotorSpeed = xSpeed + zRotation
                rightMotorSpeed = maxInput
        else:
            if zRotation >= 0.0:
                leftMotorSpeed = xSpeed + zRotation
                rightMotorSpeed = maxInput
            else:
                leftMotorSpeed = maxInput
                rightMotorSpeed = xSpeed - zRotation

        leftMotorSpeed = RobotDriveBase.limit(leftMotorSpeed) * self.maxOutput
        rightMotorSpeed = RobotDriveBase.limit(rightMotorSpeed) * self.maxOutput

        self.leftMotor.set(leftMotorSpeed)
        self.rightMotor.set(-rightMotorSpeed)

        self.feed()

    def curvatureDrive(self, xSpeed, zRotation, isQuickTurn):
        """
        Curvature drive method for differential drive platform.

        The zRotation argument controls the curvature of the robot's path rather than its rate
        of heading change. This makes the robot more controllable at high speeds. Also handles
        the robot's quick turn functionality - "quick turn" overrides constant-curvature turning
        for turn-in-place maneuvers

        :param xSpeed: The robot's speed along the X axis `[-1.0..1.0]`. Forward is positive.
        :param zRotation:  The robot's rotation rate around the Z axis `[-1.0..1.0]`. Clockwise is positive.
        :param isQuickTurn: If set, overrides constant-curvature turning for
                          turn-in-place maneuvers.
        """
        if not self.reported:
            # hal.report(hal.UsageReporting.kResourceType_RobotDrive,
            #           2,
            #           hal.UsageReporting.kRobotDrive_Curvature)
            self.reported = True

        xSpeed = RobotDriveBase.limit(xSpeed)
        xSpeed = RobotDriveBase.applyDeadband(xSpeed, self.deadband)

        zRotation = RobotDriveBase.limit(zRotation)
        zRotation = RobotDriveBase.applyDeadband(zRotation, self.deadband)

        if isQuickTurn:
            if abs(xSpeed) < self.quickStopThreshold:
                self.quickStopAccumulator = (1 - self.quickStopAlpha) * self.quickStopAccumulator + \
                                            self.quickStopAlpha * RobotDriveBase.limit(zRotation) * 2

            overPower = True
            angularPower = zRotation

        else:
            overPower = False
            angularPower = abs(xSpeed) * zRotation - self.quickStopAccumulator

            if self.quickStopAccumulator > 1:
                self.quickStopAccumulator -= 1
            elif self.quickStopAccumulator < -1:
                self.quickStopAccumulator += 1
            else:
                self.quickStopAccumulator = 0

        leftMotorSpeed = xSpeed + angularPower
        rightMotorSpeed = xSpeed - angularPower

        if overPower:
            if leftMotorSpeed > 1.0:
                rightMotorSpeed -= leftMotorSpeed - 1.0
                leftMotorSpeed = 1.0
            elif rightMotorSpeed > 1.0:
                leftMotorSpeed -= rightMotorSpeed - 1.0
                rightMotorSpeed = 1.0
            elif leftMotorSpeed < -1.0:
                rightMotorSpeed -= leftMotorSpeed + 1.0
                leftMotorSpeed = -1.0
            elif rightMotorSpeed < -1.0:
                leftMotorSpeed -= rightMotorSpeed + 1.0
                rightMotorSpeed = -1.0

        self.leftMotor.set(leftMotorSpeed * self.maxOutput)
        self.rightMotor.set(-rightMotorSpeed * self.maxOutput)

        self.feed()

    def tankDrive(self, leftSpeed, rightSpeed, squaredInputs=True):
        """Provide tank steering using the stored robot configuration.

        :param leftSpeed: The robot's left side speed along the X axis `[-1.0..1.0]`. Forward is positive.
        :param rightSpeed: The robot's right side speed along the X axis`[-1.0..1.0]`. Forward is positive.
        :param squaredInputs: If set, decreases the input sensitivity at low speeds
        """

        if not self.reported:
            hal.report(hal.UsageReporting.kResourceType_RobotDrive,
                       2,
                       hal.UsageReporting.kRobotDrive_Tank)
            self.reported = True

        leftSpeed = RobotDriveBase.limit(leftSpeed)
        leftSpeed = RobotDriveBase.applyDeadband(leftSpeed, self.deadband)

        rightSpeed = RobotDriveBase.limit(rightSpeed)
        rightSpeed = RobotDriveBase.applyDeadband(rightSpeed, self.deadband)

        # square the inputs (while preserving the sign) to increase fine
        # control while permitting full power
        if squaredInputs:
            leftSpeed = math.copysign(leftSpeed * leftSpeed, leftSpeed)
            rightSpeed = math.copysign(rightSpeed * rightSpeed, rightSpeed)

        self.leftMotor.set(leftSpeed * self.maxOutput)
        self.rightMotor.set(-rightSpeed * self.maxOutput)

        self.feed()

    def setQuickStopThreshold(self, threshold):
        """Sets the QuickStop speed threshold in curvature drive.

        QuickStop compensates for the robot's moment of inertia when stopping after a QuickTurn.

        While QuickTurn is enabled, the QuickStop accumulator takes on the rotation rate value
        outputted by the low-pass filter when the robot's speed along the X axis is below the
        threshold. When QuickTurn is disabled, the accumulator's value is applied against the computed
        angular power request to slow the robot's rotation.

        :param threshold: X speed below which quick stop accumulator will receive rotation rate values `[0..1.0]`.
        """
        self.quickStopThreshold = threshold

    def setQuickStopAlpha(self, alpha):
        """Sets the low-pass filter gain for QuickStop in curvature drive.

        The low-pass filter filters incoming rotation rate commands to smooth out high frequency
        changes.

        :param alpha: Low-pass filter gain [0.0..2.0]. Smaller values result in slower output changes.
                      Values between 1.0 and 2.0 result in output oscillation. Values below 0.0 and
                      above 2.0 are unstable.
        """
        self.quickStopAlpha = alpha

    def stopMotor(self):
        self.leftMotor.stopMotor()
        self.rightMotor.stopMotor()

        self.feed()

    def getDescription(self):
        return "Differential Drive"

    def initSendable(self, builder):
        builder.setSmartDashboardType("DifferentialDrive")
        builder.addDoubleProperty("Left Motor Speed", self.leftMotor.get, self.leftMotor.set)
        builder.addDoubleProperty(
            "Right Motor Speed",
            lambda: -self.rightMotor.get(),
            lambda x: self.rightMotor.set(-x))



# validated: 2017-12-24 DS 8b7aa61091df edu/wpi/first/wpilibj/Notifier.java
#----------------------------------------------------------------------------
# Copyright (c) FIRST 2016-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import threading

import hal

from .resource import Resource
from .robotcontroller import RobotController

class Notifier:
    
    def __init__(self, run: callable):
        
        #: The lock for the process information.
        self._processLock = threading.RLock()
        
        # Notifier handle
        self._notifier = hal.initializeNotifier()
        
        #: The time, in microseconds, at which the corresponding handler should be
        #: called. Has the same zero as Timer.getFPGATime().
        self._expirationTime = 0
        
        #: The handler passed in by the user which should be called at the
        #: appropriate interval.
        self._handler = run
        
        # Whether we are calling the handler just once or periodically.
        self._periodic = False
        
        #: If periodic, the period of the calling; if just once, stores how long it
        #: is until we call the handler.
        self._period = 0
        
        #: The thread waiting on the HAL alarm
        self._thread = threading.Thread(target=self._thread, daemon=True)
        self._thread.start()
        
        # python-specific
        Resource._add_global_resource(self)
    
    # python-specific: this is called finalize in Java
    def free(self) -> None:
        handle, self._notifier = self._notifier, None
        hal.stopNotifier(handle)
        
        # Join the thread to ensure the handler has exited.
        if self._thread.is_alive():
            # python-specific: interrupt not supported
            self._thread.join()
        
        hal.cleanNotifier(handle)

    def _updateAlarm(self) -> None:
        """Update the alarm hardware to reflect the next alarm."""
        handle = self._notifier
        if handle:
            hal.updateNotifierAlarm(handle, int(self._expirationTime * 1e6))
    
    def _thread(self) -> None:
        while True:
            notifier = self._notifier
            if not notifier:
                break
            
            curTime = hal.waitForNotifierAlarm(notifier)
            if curTime == 0:
                break
            
            with self._processLock:
                handler = self._handler
                if self._periodic:
                    self._expirationTime += self._period
                    self._updateAlarm()
            
            if handler:
                handler()
    
    def setHandler(self, handler: callable) -> None:
        """Change the handler function.
        
        :param handler: Handler
        """
        with self._processLock:
            self._handler = handler
    
    def startSingle(self, delay: float) -> None:
        """Register for single event notification. A timer event is queued for a single event after the
        specified delay.
        
        :param delay: Seconds to wait before the handler is called.
        """
        with self._processLock:
            self._periodic = False
            self._period = delay
            self._expirationTime = RobotController.getFPGATime() * 1e-6 + delay
            self._updateAlarm()
    
    def startPeriodic(self, period: float) -> None:
        """Register for periodic event notification. A timer event is queued for periodic event
        notification. Each time the interrupt occurs, the event will be immediately requeued for the
        same time interval.
        
        :param period: Period in seconds to call the handler starting one period after the call to this
        method.
        """
        with self._processLock:
            self._periodic = True
            self._period = period
            self._expirationTime = RobotController.getFPGATime() * 1e-6 + period
            self._updateAlarm()
    
    def stop(self) -> None:
        """Stop timer events from occurring. Stop any repeating timer events from occurring. This will
        also remove any single notification events from the queue. If a timer-based call to the
        registered handler is in progress, this function will block until the handler call is complete.
        """
        hal.cancelNotifierAlarm(self._notifier)



# validated: 2017-11-12 TW 21585f70a88e edu/wpi/first/wpilibj/XboxController.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import enum

import hal
from .interfaces.generichid import GenericHID

__all__ = ['XboxController']


class XboxController(GenericHID):
    """
        Handle input from Xbox 360 or Xbox One controllers connected to the Driver Station.

        This class handles Xbox input that comes from the Driver Station. Each time a value is
        requested the most recent value is returned. There is a single class instance for each controller
        and the mapping of ports to hardware buttons depends on the code in the Driver Station.
     """

    class Button(enum.IntEnum):
        kBumperLeft = 5
        kBumperRight = 6
        kStickLeft = 9
        kStickRight = 10
        kA = 1
        kB = 2
        kX = 3
        kY = 4
        kBack = 7
        kStart = 8

    def __init__(self, port: int) -> None:
        """Construct an instance of an XBoxController. The XBoxController index is the USB port on the Driver Station.

        :param port: The port on the Driver Station that the joystick is plugged into
        """
        super().__init__(port)

        hal.report(hal.UsageReporting.kResourceType_Joystick, port)

    def getX(self, hand: GenericHID.Hand) -> float:
        """Get the X axis value of the controller.

        :param hand: Side of controller whose value should be returned
        :return: The X axis value of the controller
        """
        if hand == self.Hand.kLeft:
            return self.getRawAxis(0)
        else:
            return self.getRawAxis(4)

    def getY(self, hand: GenericHID.Hand) -> float:
        """Get the Y axis value of the controller.

        :param hand: Side of controller whose value should be returned
        :return: The Y axis value of the controller
        """
        if hand == self.Hand.kLeft:
            return self.getRawAxis(1)
        else:
            return self.getRawAxis(5)

    def getTriggerAxis(self, hand: GenericHID.Hand) -> float:
        """Get the trigger axis value of the controller.

        :param hand: Side of controller whose value should be returned
        :return: The trigger axis value of the controller
        """
        if hand == self.Hand.kLeft:
            return self.getRawAxis(2)
        else:
            return self.getRawAxis(3)

    def getBumper(self, hand: GenericHID.Hand) -> bool:
        """Read the values of the bumper button on the controller.

        :param hand: Side of controller whose value should be returned.
        :return: The state of the button
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButton(self.Button.kBumperLeft)
        else:
            return self.getRawButton(self.Button.kBumperRight)

    def getBumperPressed(self, hand: GenericHID.Hand) -> bool:
        """Whether the bumper was pressed since the last check.

        :param hand: Side of controller whose value should be returned.
        :returns: Whether the button was pressed since the last check.
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButtonPressed(self.Button.kBumperLeft)
        else:
            return self.getRawButtonPressed(self.Button.kBumperRight)

    def getBumperReleased(self, hand: GenericHID.Hand) -> bool:
        """Whether the bumper was released since the last check.

        :param hand: Side of controller whose value should be returned.
        :returns: Whether the button was released since the last check.
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButtonReleased(self.Button.kBumperLeft)
        else:
            return self.getRawButtonReleased(self.Button.kBumperRight)

    def getStickButton(self, hand: GenericHID.Hand) -> bool:
        """Read the values of the stick button on the controller

        :param hand: Side of the controller whose value should be returned
        :return: The state of the button
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButton(9)
        else:
            return self.getRawButton(10)

    def getStickButtonPressed(self, hand: GenericHID.Hand) -> bool:
        """Whether the stick button was pressed since the last check.

        :param hand: Side of controller whose value should be returned.
        :return: Whether the button was pressed since the last check.
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButtonPressed(self.Button.kStickLeft)
        else:
            return self.getRawButtonPressed(self.Button.kStickRight)

    def getStickButtonReleased(self, hand: GenericHID.Hand) -> bool:
        """Whether the stick button was released since the last check.

        :param hand: Side of controller whose value should be returned.
        :return: Whether the button was released since the last check.
        """
        if hand == GenericHID.Hand.kLeft:
            return self.getRawButtonReleased(self.Button.kStickLeft)
        else:
            return self.getRawButtonReleased(self.Button.kStickRight)

    def getAButton(self) -> bool:
        """Read the value of the A button on the controller

        :return: The state of the A button
        """
        return self.getRawButton(self.Button.kA)

    def getAButtonPressed(self) -> bool:
        """Whether the A button was pressed since the last check.

        :returns: Whether the button was pressed since the last check.
        """
        return self.getRawButtonPressed(self.Button.kA)

    def getAButtonReleased(self) -> bool:
        """Whether the A button was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kA)

    def getBButton(self) -> bool:
        """Read the value of the B button on the controller

        :return: The state of the B button
        """
        return self.getRawButton(self.Button.kB)

    def getBButtonPressed(self) -> bool:
        """Whether the B button was pressed since the last check.

        :returns: Whether the button was pressed since the last check.
        """
        return self.getRawButtonPressed(self.Button.kB)

    def getBButtonReleased(self) -> bool:
        """Whether the B button was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kB)

    def getXButton(self) -> bool:
        """Read the value of the X button on the controller

        :return: The state of the X button
        """
        return self.getRawButton(self.Button.kX)

    def getXButtonPressed(self) -> bool:
        """Whether the X button was pressed since the last check.

        :returns: Whether the button was pressed since the last check.
        """
        return self.getRawButtonPressed(self.Button.kX)

    def getXButtonReleased(self) -> bool:
        """Whether the X button was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kX)

    def getYButton(self) -> bool:
        """Read the value of the Y button on the controller

        :return: The state of the Y button
        """
        return self.getRawButton(self.Button.kY)

    def getYButtonPressed(self) -> bool:
        """Whether the Y button was pressed since the last check.

        :returns: Whether the button was pressed since the last check.
        """
        return self.getRawButtonPressed(self.Button.kY)

    def getYButtonReleased(self) -> bool:
        """Whether the Y button was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kY)

    def getBackButton(self) -> bool:
        """Read the value of the Back button on the controller

        :return: The state of the Back button
        """
        return self.getRawButton(self.Button.kBack)

    def getBackButtonPressed(self) -> bool:
        """Whether the Back button was pressed since the last check.

        :returns: Whether the button was pressed since the last check.
        """
        return self.getRawButtonPressed(self.Button.kBack)

    def getBackButtonReleased(self) -> bool:
        """Whether the Back button was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kBack)

    def getStartButton(self) -> bool:
        """Read the value of the Start button on the controller

        :return: The state of the Start button
        """
        return self.getRawButton(self.Button.kStart)

    def getStartButtonPressed(self) -> bool:
        """Whether the Start button was pressed since the last check.

        :returns: Whether the button was pressed since the last check.
        """
        return self.getRawButtonPressed(self.Button.kStart)

    def getStartButtonReleased(self) -> bool:
        """Whether the Start button was released since the last check.

        :returns: Whether the button was released since the last check.
        """
        return self.getRawButtonReleased(self.Button.kStart)
